﻿﻿package org.fuseproject.go {
	import flash.display.Sprite;
	import flash.events.Event;
	import flash.events.TimerEvent;
	import flash.utils.Dictionary;
	import flash.utils.Timer;
	import flash.utils.getTimer;
	
	import org.fuseproject.go.interfaces.IHandlerManager;
	import org.fuseproject.go.interfaces.IPropertyHandler;
	import org.fuseproject.go.interfaces.IUpdatable;
	
	/**
	 * Provides <code>update</code> calls to <code>IUpdatable</code> instances on their specified <code>pulseInterval</code>.
	 * 
	 * <p>Go is a generic set of base classes for creating AS3 animation tools.
	 * It provides structure and core functionality, but does not define the
	 * specifics of animation-handling classes like tweens.</p>
	 * 
	 * <p><i>Important: Store your custom Go classes in a package bearing your 
	 * own classpath, not in the core package! This will help avoid confusion 
	 * with other authors' work.</i></p>
	 * 
	 * <p>GoEngine sits at the center of the Go system, and along with the IUpdatable 
	 * interface is the only required element for using Go. GoEngine references two other 
	 * interfaces for adding system-wide managers, IHandlerManager and IPropertyHandler.
	 * All other classes in the go package are merely one suggestion of how a 
	 * system could be structured within Go, and may be considered optional 
	 * elements. To create an API using the provided classes, you simply need 
	 * to extend the item classes LinearGo and PhysicsGo to create animation items. 
	 * </p>
	 * <p>
	 * GoEngine serves two purposes: first, it keeps a large system efficient 
	 * by stacking and running updates on blocks of items. Note that any IUpdatable 
	 * instance may specify its own pulseInterval; items with matching pulses 
	 * are grouped into queues for efficiency. Its second purpose is centralization. 
	 * By using a single hub for pulse-driven items of all types, management classes 
	 * can be attached to GoEngine to run processes across items. This is done voluntarily 
	 * by the end-user with <code>addManager</code>, which keeps management entirely 
	 * compile-optional and extensible.
	 * </p>
	 * <p>You may modify any class in the go package to suit your project's needs.</p>
	 *  
	 * Go is a community initiative led by Moses Gunesch at MosesSupposes.com. Please
	 * visit the <a href="http://go.mosessupposes.com/">Go blog</a> for more information.
	 * 
	 * <p></i>{In the game of Go, the wooden playing board, or Goban, features a grid
	 *  on which black & white go-ishi stones are laid at its intersections.}</i></p>
	 * 
	 * @author Moses Gunesch / mosessupposes.com (c) all rights reserved
	 */
	public class GoEngine 
	{
		// -== Constants ==-
		
		public static const INFO:String = "Go 0.1.1 (c) MosesSupposes.com";
		
		// -== Settable Class Defaults ==-
		
		/**
		 * A pulseInterval that runs on the player's natural framerate, 
		 * which is often most efficient.
		 */
		public static const ENTER_FRAME	: int = -1;

		// -== Protected Properties ==-
		
		private static var items : Dictionary = new Dictionary(false);
		private static var timers : Dictionary = new Dictionary(false);
		private static var timerCounts : Dictionary = new Dictionary(false);
		private static var pulseSprite : Sprite;
		private static var paused : Boolean = false;
		private static var itemCount : uint = 0;
		private static var managers : Dictionary = new Dictionary(false);
		
		// -== Public Class Methods ==-
		
		/**
		 * Enables the extending of this class' functionality with a tight
		 * coupling to an IHandlerManager. (Tight coupling is crucial in 
		 * such a time-sensitive context; standard events are too asynchronous.) 
		 * All items that implement IPropertyHandler are reported to registered
		 * managers as they add and remove themselves from GoEngine.
		 * 
		 * @see removeManager
		 */
		public static function addManager( instance:IHandlerManager ):void
		{
			managers[ instance ] = 1; // Dictionary avoids duplication
		}
		
		/**
		 * Unregisters any manager set in <code>addManager</code>.
		 * 
		 * @see addManager
		 */
		public static function removeManager( instance:IHandlerManager ):void
		{
			delete managers[ instance ];
		}
		
		/**
		 * Adds an IUpdatable instance to an update-queue corresponding to
		 * the item's pulseInterval property.
		 * 
		 * @param item		Any object implementing IUpdatable that wishes 
		 * 					to receive update calls on a pulse.
		 * 					
		 * @return			Success
		 */
		public static function addItem( item:IUpdatable ):Boolean
		{
			try {
				if (!items[ item.pulseInterval ]) {
					// Create a itemlist for each pulse. Grouping items in a list keeps update cycles efficient.
					items[ item.pulseInterval ] = new Array();
				}
				(items[ item.pulseInterval ] as Array).push(item);
				addPulse(item.pulseInterval);
				
				// Report IPropertyHandler instances to registered managers
				if (item is IPropertyHandler) {
					for (var manager:Object in managers)
						(manager as IHandlerManager).reserve( item as IPropertyHandler );
				}
				itemCount++;
				return true;
			}
			catch (e:Error) {
				trace(e.message);
			}
			return false;
		}
		
		/**
		 * Removes an item from the queue and removes its pulse timer if
		 * the queue is depleted.
		 * 
		 * @param item		Any IUpdatable previously added that wishes 
		 * 					to stop receiving update calls.
		 * 					
		 * @return			Success
		 */
		public static function removeItem( item:IUpdatable ):Boolean
		{
			try {
				var list:Array = (items[ item.pulseInterval ] as Array);
				var i:int = list.indexOf(item);
				if (i>-1)
					list.splice( i, 1 );
				GoEngine.removePulse(item.pulseInterval);
				
				// Report IPropertyHandler item removal to registered managers.
				if (item is IPropertyHandler) {
					for (var manager:Object in managers)
						(manager as IHandlerManager).release( item as IPropertyHandler );
				}
				itemCount--;
				return true;
			}
			catch (e:Error) {
				trace(e.message);
			}
			return false;
		}
		
		/**
		 * Removes all items and resets the engine, 
		 * or removes just items running on a specific pulse.
		 * 
		 * @param pulseInterval		Optionally filter by a specific pulse 
		 * 							such as ENTER_FRAME or a number of milliseconds.
		 */
		public static function clear(pulseInterval:Number = NaN) : void
		{
			for (var pulse:Object in items) {
				if (!pulseInterval || pulseInterval==pulse) {
					for (var item:Object in items)
						removeItem(item as IUpdatable);
				}
			}
		}
		
		/**
		 * Retrieves number of active items in the engine 
		 * or active items running on a specific pulse.
		 * 
		 * @param pulseInterval		Optionally filter by a specific pulseInterval
		 *							such as ENTER_FRAME or a number of milliseconds.
		 * 
		 * @return					Number of active items in the Engine.
		 */
		public static function getCount(pulseInterval:Number = NaN) : uint
		{
			if (isNaN(pulseInterval)) {
				return itemCount;
			}
			try {
				var count:Number = 0;
				for (var item:Object in items[ pulseInterval ])
					count++;
				return count;
			}
			catch (e:Error) {
				trace(e.message);
			}
			return 0;
		}
		
		/**
		 * @return			The paused state of engine.
		 */
		public static function getPaused() : Boolean {
			return paused;
		}
		
		/**
		 * Pauses or resumes all animation globally by suspending processing.
		 * 
		 * @param pause		Pass false to resume if currently paused.
		 */
		public static function setPaused(pause:Boolean=true) : void
		{
			if (paused==pause) return;
			paused = pause;
			for (var pulse:Object in timers) {
				if (pulse==ENTER_FRAME) {
					if (paused)
						pulseSprite.removeEventListener(Event.ENTER_FRAME, update);
					else
						pulseSprite.addEventListener(Event.ENTER_FRAME, update);
				}
				else {
					if (paused)
						Timer( timers[ pulse ] ).stop();
					else
						Timer( timers[ pulse ] ).start();
				}
			}
		}
		
		// -== Private Class Methods ==-
		
		/**
		 * Executes the update queue corresponding to the dispatcher's interval.
		 * 
		 * @param event			TimerEvent or Sprite ENTER_FRAME Event
		 */
		private static function update(event:Event) : void 
		{
			var currentTime:Number = getTimer();
			var pulse:int = (event is TimerEvent ? ( event.target as Timer ).delay : ENTER_FRAME);
			var list:Array = (items[ pulse ] as Array);
			for each (var item:IUpdatable in list) {
				item.update(currentTime);
			}
		}
		
		/**
		 * Creates new timers when a previously unused interval is specified,
		 * and tracks the number of items associated with that interval.
		 * 
		 * @param pulse			The pulseInterval requested
		 */
		private static function addPulse(pulse : int) : void
		{
			if (!timerCounts[ pulse ]) {
				timerCounts[ pulse ] = 0;
				if (pulse==ENTER_FRAME) {
					pulseSprite = new Sprite();
					pulseSprite.addEventListener(Event.ENTER_FRAME, update);
				}
				else {
					timers[ pulse ] = new Timer(pulse);
					Timer( timers[ pulse ] ).addEventListener(TimerEvent.TIMER, update);
					Timer( timers[ pulse ] ).start();
				}
			}
			timerCounts[ pulse ] += 1;
		}
		
		/**
		 * Tracks whether a removed item was the last one using a timer 
		 * and if so, removes that timer.
		 * 
		 * @param pulse			The pulseInterval corresponding to an item being removed.
		 */
		private static function removePulse(pulse : int) : void
		{
			if (--timerCounts[ pulse ] == 0) {
				if (pulse==ENTER_FRAME) {
					pulseSprite.removeEventListener(Event.ENTER_FRAME, update);
					pulseSprite = null;
				}
				else {
					Timer( timers[ pulse ] ).stop();
					Timer( timers[ pulse ] ).removeEventListener(TimerEvent.TIMER, update);
				}
				delete timers[ pulse ];
				delete timerCounts[ pulse ];
			}
		}
	}
}
