
package org.fuseproject.go {
	import flash.events.EventDispatcher;
	import flash.utils.getQualifiedClassName;
	
	import org.fuseproject.go.interfaces.IPlayable;

	/**
	 * Top-level abstract base class for IPlayable instances.
	 * 
	 * <p>
	 * This class fleshes out the IPlayable interface with a standard set of 
	 * play-state constants that can be used throughout the Go system. It also adds 
	 * a convenient <code>id</code> property that can be used to more easily track 
	 * instances, or even label them since it public and untyped. On its own, this 
	 * class does nothing but set its own <code>state</code>.
	 * </p>
	 * 
	 * @author Moses Gunesch / mosessupposes.com (c) all rights reserved
	 */
	public class PlayableBase extends EventDispatcher implements IPlayable {
		
		// -== Play-state Constants ==-
		
		/**
		 * Instance play is currently stopped.
		 */
		public static const STOPPED			: String = "STOPPED";
		
		/**
		 * Instance play is currently paused.
		 */
		public static const PAUSED			: String = "PAUSED";
		
		/**
		 * Instance is currently playing a delay, but has not started playing.
		 * Delays are a non-universal feature that must be custom-implemented,
		 * so some subclasses of PlayableBase may never use this constant. 
		 */
		public static const PLAYING_DELAY	: String = "PLAYING_DELAY";
		
		/**
		 * Instance play is currently playing.
		 */
		public static const PLAYING			: String = "PLAYING";
		
		/**
		 * Subclasses must dispatch GoEvent.COMPLETE and/or GoEvent.STOP to be 
		 * usable.
		 */
		public static const directUseError: String = "Direct use of PlayableBase is not allowed, use subclasses only.";
		
		// -== Public Properties ==-
		
		/**
		 * An arbitrary id value for the convenient identification of any
		 * instance. Can be set to an instance-count number by Go classes. 
		 */
		public var id:*;

		/**
		 * Returns the value of one of this class' play-state constants.
		 * @see STOPPED
		 * @see PAUSED
		 * @see PLAYING_DELAY
		 * @see PLAYING
		 */
		public function get state() : String {
			return _state;
		}

		// -== Protected Properties ==-
		
		private static var _idCounter	: int = -1;
		protected var _state : String = STOPPED;
		
		// -== Public Methods ==-
		
		public function PlayableBase() : void {
			var className:String = getQualifiedClassName(this);
			if (className.slice(className.lastIndexOf("::")+2) == "PlayableBase") {
				throw new Error(directUseError);
			}
			id = ++ _idCounter;
		}
		
		public function start() : Boolean {
			_state = PLAYING;
			return true;
		}
		
		public function stop() : Boolean {
			if (_state==STOPPED)
				return false;
			_state = STOPPED;
			return true;
		}
		
		public function pause() : Boolean {
			if (_state==STOPPED || _state==PAUSED)
				return false;
			_state = PAUSED;
			return true;
		}
		
		public function resume() : Boolean {
			if (_state != PAUSED)
				return false;
			_state = PLAYING;
			return true;
		}
		
		public function skipTo(position : Number) : Boolean {
			_state = PLAYING;
			return true;
		}
		
		/**
		 * Appends the regular toString value with the instance's id.
		 * 
		 * @return	String representation of this instance.
		 */
		override public function toString():String {
			var s:String = super.toString();
			var addLast:Boolean = (s.charAt(s.length-1)=="]");
			if (addLast) s = s.slice(0,-1);
			if (id is String) s += " id:\"" + id + "\"";
			else s += " id:" + id;
			if (addLast) s += "]";
			return s;
		}
	}
}
