package org.fuseproject.go.items {
	import org.fuseproject.go.GoEngine;
	import org.fuseproject.go.PlayableBase;

	/**
	 * Abstract base animation class for other base classes like LinearGo and PhysicsGo.
	 * 
	 * <p> 
	 * This class extends PlayableBase to add features common to any animation item, 
	 * either linear or physics (LinearGo and PhysicsGo both extend this class).
	 * Animation items add themselves to GoEngine to run on a pulse, so the IUpdatable
	 * interface is implemented here, although the update method needs to be subclassed.
	 * </p>
	 * <p> 
	 * Animation items should individually implement the standard <code>useRounding</code>
	 * and <code>useRelative</code> options. Three user-accessible class default settings
	 * are provided for those and <code>pulseInterval</code>, while play-state constants
	 * live in the superclass PlayableBase.
	 * </p>
	 * 
	 * @author Moses Gunesch / mosessupposes.com (c) all rights reserved
	 */
	public class GoItem extends PlayableBase
	{
		// -== Settable Class Defaults ==-
		
		/**
		 * Class default for the instance property <code>pulseInterval</code>.
		 * @see pulseInterval
		 */
		public static var defaultPulseInterval : Number = GoEngine.ENTER_FRAME;
		
		/**
		 * Class default for the instance property <code>useRounding</code>.
		 * @see useRounding
		 */
		public static var defaultUseRounding : Boolean = false;
		
		/**
		 * Class default for the instance property <code>useRelative</code>.
		 * @see useRelative
		 */
		public static var defaultUseRelative : Boolean = false;
		
		/**
		 * Alters the play speed for instances of any subclass that factors 
		 * this value into its calculations, such as LinearGo.
		 * 
		 * <p>A setting of 2 should result in half-speed animations, while a setting
		 * of .5 should double animation speed. Note that changing this property at 
		 * runtime does not usually affect already-playing items.</p>
		 */
		public static var timeMultiplier : Number = 1;
		
		// -== Public Properties ==-
		
		/**
		 * Required by IUpdatable. Defines the pulse on which <code>update</code> is called.
		 * 
		 * <p> 
		 * Can be a number of milliseconds for Timer-based updates or 
		 * <code>GoEngine.ENTER_FRAME</code> (-1) for updates synced to the 
		 * Flash Player's framerate. If not set manually, the class 
		 * default <code>defaultPulseInterval</code> is adopted.
		 * </p>
		 * 
		 * @see defaultPulseInterval
		 * @see org.fuseproject.go.linear.GoEngine#ENTER_FRAME
		 */
		public function get pulseInterval() : int {
			return _pulse;
		}
		public function set pulseInterval(interval:int) : void {
			if (_state==STOPPED && (interval >= 0 || interval==GoEngine.ENTER_FRAME)) {
				_pulse = interval;
			}
		}
		
		/**
		 * Subclasses should use the <code>correctValue</code> method on
		 * all calculated values before applying them to targets,
		 * which corrects NaN's to 0 and applies Math.round if this
		 * property is set to true.
		 * 
		 * @see correctValue
		 */
		public var useRounding			: Boolean = defaultUseRounding;
		
		/**
		 * IMPORTANT: <i>All items need to implement this functionality 
		 * individually</i>. When true, user-set values should be calculated 
		 * as relative to their existing value ("from" vs. "to"), if possible. 
		 * 
		 * <p>
		 * See an example in the documentation for <code>LinearGo.start</code>.
		 * </p>
		 * <p>
		 * Items that handle more than one property at once, such as a bezier
		 * curve, might want to implement a useRelative option for each property 
		 * instead of using this overall item property, which is included here
		 * to define a convention for standard single-property items.
		 * </p>
		 * 
		 * @see defaultUseRelative
		 */
		public var useRelative			: Boolean = defaultUseRelative;
		
		
		// -== Protected Properties ==-
		
		protected var _pulse			: int = defaultPulseInterval;
		
		// -== Public Methods ==-
		
		/**
		 * Constructor.
		 */
		public function GoItem() {
			super();
		}
		
		/**
		 * IMPORTANT: <i>Subclasses need to implement this functionality 
		 * individually</i>. When updating animation targets, always call
		 * <code>correctValue</code> on results first. This corrects any
		 * NaN's to 0 and applies Math.round if <code>useRounding</code> 
		 * is active.
		 * 
		 * <p>For example, a LinearGo <code>onUpdate</code> method might contain:</p>
		 * <pre>
		 * target[ prop ] = correctValue(start + (change * _position));
		 * </pre> 
		 * 
		 * @see useRounding
		 * @see defaultUseRounding
		 */
		public function correctValue(value:Number):Number 
		{
			if (isNaN(value)) 
				return 0;
			
			if (useRounding) 
				return Math.round(value);
			
			return value;
		}
		
		/**
		 * Required by IUpdatable. Perform updates on a pulse.
		 * 
		 * @param currentTime	A clock time that should be used instead of getTimer
		 * 						in performing update calculations. (The value is usually 
		 * 						not more than a couple milliseconds different than getTimer
		 * 						but using it tightly syncs all items in the timer group
		 * 						and can make a perceptible difference.)
		 */
		public function update(currentTime : Number) : void {
			// override this method.
		}
	}
}
