﻿package org.fuseproject.go.items {
	import flash.utils.getTimer;
	
	import org.fuseproject.go.GoEngine;
	import org.fuseproject.go.events.GoEvent;
	import org.fuseproject.go.interfaces.IUpdatable;	

	/**
	 * Dispatched just prior to an animation's first update after the delay 
	 * has completed, if one was set. Any number of callbacks may also be 
	 * associated with this cycle using <code>addCallback</code>.
	 * @eventType org.fuseproject.go.events.START
	 */
	[Event(name="START", type="org.fuseproject.go.events.GoEvent")]

	/**
	 * Dispatched on the animation's update pulse. Any number of callbacks
	 * may also be associated with this cycle using <code>addCallback</code>.
	 * @eventType org.fuseproject.go.events.UPDATE
	 */
	[Event(name="UPDATE", type="org.fuseproject.go.events.GoEvent")]

	/**
	 * Dispatched if an animation is manually stopped. Any number of callbacks
	 * may also be associated with this cycle using <code>addCallback</code>.
	 * @eventType org.fuseproject.go.events.STOP
	 */
	[Event(name="STOP", type="org.fuseproject.go.events.GoEvent")]

	/**
	 * Dispatched on an animation's final update, just after the last update event.
	 * Any number of callbacks may also be associated with this cycle using <code>addCallback</code>.
	 * @eventType org.fuseproject.go.events.COMPLETE
	 */
	[Event(name="COMPLETE", type="org.fuseproject.go.events.GoEvent")]

	/**
	 * LinearGo extends the base class GoItem to define a playable A-to-B animation. 
	 * 
	 * <p>
	 * <b>LinearGo: A very simple tween</b>
	 * </p>
	 * <p>
	 * A LinearGo instance is a playable object that animates a single number. It dispatches events 
	 * and callbacks associated with the animation's start, update and completion. Instances can be used 
	 * directly, or easily subclassed to build custom tweening APIs. LinearGo extends GoItem, which 
	 * provides basic settings shared by physics and tween items. These include play-state contants and 
	 * a <code>state</code> property, <code>pulseInterval</code>, and the two common animation options 
	 * <code>useRounding</code> and <code>useRelative</code>.
	 * </p>
	 * <p>
	 * The tween can be customized using the instance properties <code>duration</code>, <code>easing</code> 
	 * and <code>delay</code>. The number crunched by a LinearGo is readable in its <code>position</code> 
	 * property. This number always starts at 0 and completes at 1, regardless of the tween's duration 
	 * or easing (those parameters are factored in to produce accurate fractional in-between values).  
	 * As the tween runs, you can use <code>position</code> as a multiplier to animate virtually anything: 
	 * motion, alpha, a sound level, the values in a ColorTransform, BitmapFilter, a 3D scene, and so on.
	 * Note that at times position may be less than 0 or greater than 1 depending on the easing function.
	 * </p>
	 * <p>
	 * The START event occurs just before the first update (after the delay). UPDATE is fired once on 
	 * <i>every</i> update pulse, and COMPLETE just after the final update. The STOP event, more common in 
	 * physics animation, is fired by LinearGo only if a tween is stopped before it completes. Besides using
	 * standard events, you can store callback functions (method-closures) using <code>addCallback</code>.
	 * Any number of callbacks can be associated with each GoEvent type. This alternative to the standard event 
	 * model was included in LinearGo since it's a common feature of many modern tweening APIs, and very 
	 * slightly more efficient than standard events. 
	 * </p>
	 * <p>
	 * <b>Subclassing to create custom tweens</b>
	 * </p>
	 * <p>
	 * <i>Important: Store your custom tween classes in a package bearing your own classpath, not in the core 
	 * package! This will help avoid confusion with other authors' work.</i> 
	 * </p>
	 * <p>
	 * It's possible to build virtually 
	 * any tweening API over LinearGo because all of the specifics are left up to you: target objects, tweenable 
	 * properties, tween values — and importantly, the datatypes of all of these. 
	 * </p>
	 * <p>
	 * A basic subclass can be created in three steps: Gathering target & property information, subclassing the 
	 * <code>start</code> method to set up the tween, and finally subclassing the <code>onUpdate</code> method 
	 * to affect the tween. The first step, gathering tween target and property information, can be done by 
	 * writing getter/setter properties, customizing the constructor, or both. Consider various options such as 
	 * allowing for single vs. multiple target objects, open vs. specific tween properties, and so on. The next 
	 * step, subclassing <code>start</code>, involves figuring the tween's amount of change. In doing so you need
	 * to factor for the standard GoItem option <code>useRelative</code>, which should allow the user to declare 
	 * tween values as relative to existing values, instead of fixed absolutes. In the final step, you subclass 
	 * <code>onUpdate</code> to apply the tween, using the <code>_position</code> calculated by this base class:
	 * 
	 * <pre>target[ propName ] = super.correctValue(start + change * _position);</pre>
	 * 
	 * The helper method <code>correctValue</code> is provided in the superclass GoItem, to clean up NaN values 
	 * and apply rounding when <code>useRounding</code> is activated. That's it — events and callbacks are 
	 * dispatched by LinearGo, so subclasses can remain simple.
	 * </p>
	 * <p>
	 * An optional fourth step will make your custom tween compatible with Go managers. To do this, implement 
	 * the IPropertyHandler interface. (GoOverlapMonitor prevents different tween instances from handling the same 
	 * property at once; you can build other managers as well.)
	 * </p>
	 * 
	 * {In the game of Go a black or white stone is called a go-ishi.}
	 * 
	 * @author Moses Gunesch / mosessupposes.com (c) all rights reserved
	 */
	public class LinearGo extends GoItem implements IUpdatable
	{
		// -== Settable Class Defaults ==-
		
		/**
		 * Class default for the instance property delay.
		 * @see delay
		 */
		public static var defaultDelay : Number = 0;
		
		/**
		 * Class default for the instance property duration.
		 * @see duration
		 */
		public static var defaultDuration : Number = 1;
		
		/**
		 * Class default for the instance property easing. 
		 * <p>(This default is fl.motion.easing.Quintic.easeOut.)</p>
		 * @see easing
		 */
		public static var defaultEasing:Function = function(t:Number, b:Number, c:Number, d:Number) : Number {
			return c * ((t = t / d - 1) * t * t * t * t + 1) + b;
		};
		
		// -== Pulic Properties ==-
		
		/**
		 * Number of seconds after start() call that the LinearGo begins processing.
		 * <p>If not set manually, the class default defaultDelay is adopted.</p>
		 * @see defaultDelay
		 */
		public function get delay():Number {
			return _delay;
		}
		public function set delay(seconds:Number):void {
			if (_state==STOPPED && seconds >= 0) {
				_delay = seconds;
			}
		}
		
		/**
		 * Number of seconds the LinearGo takes to process.
		 * <p>If not set manually, the class default defaultDuration is adopted.</p>
		 * @see defaultDuration
		 */
		public function get duration():Number {
			return _duration;
		}
		public function set duration(seconds:Number):void {
			if (_state==STOPPED && seconds >= 0) {
				_duration = seconds;
			}
		}
		
		/**
		 * Any standard easing-equation function such as the ones found in
		 * the Flash package fl.motion.easing or the flex package mx.effects.easing.
		 * 
		 * <p>If not set manually, the class default defaultEasing is adopted. An error
		 * is thrown if the function does not follow the typical format. For easings 
		 * that accept more than four parameters use <code>extraEasingParams</code>.
		 * </p>
		 * 
		 * @see defaultEasing
		 * @see extraEasingParams
		 */
		public function get easing():Function {
			return _easing;
		}
		public function set easing(type:Function):void {
			if (_state==STOPPED) {
				try {
					if (type(1,1,1,1) is Number) {
						_easing = type;
					}
				}
				catch (e:Error) {
					throw e;
				}
			}
		}
		
		/**
		 * Additional parameters to pass to easing functions that accept more than four.
		 * @see easing
		 */
		public function get extraEasingParams() : Array {
			return _extraEaseParams;
		}
		public function set extraEasingParams(params:Array):void {
			if (_state==STOPPED && params is Array && params.length>0) {
				_extraEaseParams = params;
				_hasParams = true;
			}
		}
		
		//TODO:Implement cycles option in complexUpdate
		/**
		 * (This feature has not been written yet.)
		 * <p>
		 * Makes a Go cycle back and forth between start and end. The
		 * default value is 1 meaning tween from start to end; if set to
		 * 2 or higher the tween returns, replays, and so on; if set to
		 * 0 or INFINITY the tween cycles indefinitely until stop() is called.
		 * </p>
		 * 
		 * <p>
		 * For cycles other than 1 you may set an alternate easing for 
		 * return-cycles using easingOnCycle.
		 * </p>
		 * @see easingOnCycle
		 */
		public function get cycles():uint {
			return _cycles;
		}
		public function set cycles(value:uint):void {
			if (_state==STOPPED) {
				_cycles = value;
				setUpdateFunction();
			}
		}
		
		/**
		 * (This feature has not been written yet.)
		 * 
		 * <p>
		 * Optionally specify a different easing for each alternating cycle, if the
		 * cycles property is set to a value other than 1. For example an easing:easeOut
		 * with an easingOnCycle:easeIn will produce a more natural-looking result.
		 * </p>
		 * 
		 * <p>
		 * A standard easing-equation function such as the ones found in
		 * the Flash package fl.motion.easing or the flex package mx.effects.easing.
		 * If not set manually, the class default defaultEasing is adopted. Note that an
		 * error is thrown if the function does not follow the typical format.
		 * </p>
		 * @see cycles
		 * @see defaultEasing
		 * @see easing
		 * @see extraEasingParams
		 */
		public function get easingOnCycle():Function {
			return _easing2;
		}
		public function set easingOnCycle(type:Function):void {
			if (_state==STOPPED) {
				try {
					if (type(1,1,1,1) is Number) {
						_easing2 = type;
					}
				}
				catch (e:Error) {
					throw e;
				}
			}
		}
		
		//TODO:Implement useFrames option in complexUpdate
		/**
		 * (This feature has not been written yet.)
		 */
		public function get useFrames():Boolean {
			return _useFrames;
		}
		public function set useFrames(value:Boolean):void {
			if (_state==STOPPED) {
				_useFrames = value;
				setUpdateFunction();
			}
		}		
		
		/**
		 * A number between 0 and 1 representing the current tween value.
		 * 
		 * <p>Use this number as a multiplier to apply values to targets 
		 * across time.<p> 
		 * 
		 * <p>Here's an example of what an overridden update method might contain:</p>
		 * <pre>
		 * super.update(currentTime);
		 * target[ propName ] = super.correctValue(startValue + change*_position);
		 * </pre> 
		 */
		public function get position():Number {
			return _position;
		}
		
		
		// -== Protected Properties ==-
		
		protected var _delay 			: Number = defaultDelay;
		protected var _duration 		: Number = defaultDuration;
		protected var _tweenDuration	: Number;
		protected var _easing 			: Function = defaultEasing;
		protected var _easeParams		: Array;
		protected var _extraEaseParams	: Array;
		protected var _hasParams		: Boolean = false;
		protected var _cycles 			: uint = 1;
		protected var _easing2	 		: Function;
		protected var _currentFrame		: int;
		protected var _position			: Number;
		protected var _startTime		: Number;
		protected var _endTime 			: Number;
		protected var _pauseTime 		: Number;
		protected var _useFrames		: Boolean = false;
		protected var _doComplexUpdate	: Boolean = false;
		protected var _callbacks		: Object;
		protected var _updateEvent		: GoEvent;
		
		// -== Public Methods ==-
		
		/**
		 * Constructor. 
		 * 
		 * <p>The inputs here should not be used as a convention. Instead,
		 * define the constructors that fit your tween classes best.</p>
		 */
		public function LinearGo(	delay	 			: Number=NaN,
									duration 			: Number=NaN,
									easing 				: Function=null,
									extraEasingParams	: Array=null,
									useRounding			: Boolean=false,
									pulseInterval		: Number=NaN ) {
			if (delay) this._delay = delay;
			if (duration) this._duration = duration;
			if (easing!=null) this.easing = easing;
			if (extraEasingParams) this._extraEaseParams = extraEasingParams;
			if (useRounding) this.useRounding = useRounding;
			if (pulseInterval) this._pulse = pulseInterval;
		}
		
		/**
		 * Starts play for this LinearGo instance using GoEngine.
		 * 
		 * <p>
		 * This method should always be subclassed to handle the 
		 * <code>useRelative</code> feature of LinearGo. If <code>useRelative</code> 
		 * is true, calculate tween values relative to the target object's 
		 * existing value as in the example below.
		 * </p>
		 * <p>
		 * It's most convenient for update methods if you also store the
		 * target's existing start value, and precalculate the change
		 * between start and end so you don't have to do that each update.
		 * </p>
		 * 
		 * <pre>
		 * protected var _target : DisplayObject;
		 * protected var _width : Number;
		 * protected var _changeWidth : Number;
		 * 
		 * override public function start():Boolean 
		 * {
		 *     if (!_target || !_width || isNaN(_width))
		 *         return false;
		 * 
		 *     _startWidth = _target.width;
		 * 
		 *     if (useRelative) {
		 *         _changeWidth = _width;
		 *     } else {
		 *         _changeWidth = (_width - _startWidth);
		 *     }
		 *     
		 *     return (super.start());
		 * }
		 * </pre>
		 * 
		 * @return Successful addition of the item to GoEngine
		 */
		override public function start() : Boolean {
			stop(); // does nothing if already stopped.
			if (GoEngine.addItem(this)==false)
				return false;
			super.start();
			if (_delay > 0)
				_state = PLAYING_DELAY; // has to be set here since delay is not included in PlayableBase.
			return true;
		}
		
		/**
		 * Ends play for this LinearGo instance and dispatches a GoEvent.STOP
		 * event if the tween is incomplete. This method does not typically 
		 * require subclassing.
		 * 
		 * @return Successful removal of the item from GoEngine
		 */
		override public function stop() : Boolean {
			if (super.stop()==false)
				return false;
			var completed:Boolean = (_position==1);
			_position = 0;
			_currentFrame = 0;
			_pauseTime = NaN;
			_startTime = NaN;
			var success:Boolean = GoEngine.removeItem(this);
			if (!completed) // otherwise a COMPLETE event was dispatched.
				dispatch( GoEvent.STOP );
			return success;
		}
		
		/**
		 * Pauses play (including delay) for this LinearGo instance.
		 * This method does not typically require subclassing.
		 * 
		 * @return Success
		 * @see resume
		 * @see org.fuseproject.go.linear.GoEngine#setPaused 
		 */
		override public function pause() : Boolean {
			if (super.pause()==false)
				return false;
			//TODO: Test special case:if start then pause are called back to back, pause hits before setStartTime. Does this mess anything up later?
			_pauseTime = getTimer();
			return true;
		}
		
		/**
		 * Resumes previously paused play, including delay.
		 * This method does not typically require subclassing.
		 * 
		 * @return Success
		 * @see pause
		 * @see org.fuseproject.go.linear.GoEngine#setPaused
		 */
		override public function resume() : Boolean {
			if (super.resume()==false)
				return false;
			var currentTime:Number = getTimer();
			setStartTime(currentTime - (_pauseTime - _startTime)); 
			_pauseTime = NaN;
			if (_startTime > currentTime)
				_state = PLAYING_DELAY;
			return true;
		}
		
// TODO: should skipTo include _delay - or an extra parameter for includeDelay? Good: if skipping a group, it's better to maintain delays. Bad: cycles - there is no delay after the first cycle so the limit changes for the user. Add a parameter to skipTo?
		/**
		 * Skips to a point within the tween, ignoring delay, and plays. 
		 * This method does not typically require subclassing.
		 * 
		 * <p>
		 * Rewind by passing 0, or fast-forward by passing the tween's 
		 * full duration (or any greater value). For a 3-second tween, 
		 * skipTo(2) will discard the delay if it's running, and start 
		 * playing the tween 2/3 of the way through. Note that even if
		 * you've set <code>timeMultiplier</code>, you should pass a 
		 * time value based on the tween's original duration.
		 * </p>
		 * 
		 * @param seconds	A number between zero and the tween's duration.
		 * @return Success
		 */
		override public function skipTo(seconds : Number) : Boolean 
		{
			if (_state==PAUSED)
				_pauseTime = NaN;
			else if (_state==STOPPED)
				start();
			super.skipTo(seconds);
			if (seconds < 0 || isNaN(seconds)) 
				seconds = 0;
			var currentTime:Number = getTimer();
			setStartTime( currentTime - (Math.min(seconds, _duration/*+_delay*/) * 1000 * Math.max(0, timeMultiplier)) );
			update(currentTime);
			return true;
		}
		
		/**
		 * An alternative to subscribing to events is to store callbacks. You can 
		 * associate any number of callbacks with the primary GoEvent types START,
		 * UPDATE, COMPLETE, and STOP (only fired if the tween is stopped before it 
		 * completes).
		 * 
		 * <p>
		 * Note that there is little difference between using callbacks and events.
		 * Both are common techniques used in many various modern tweening APIs. Callbacks 
		 * are slightly faster, but this won't normally be noticeable unless thousands of 
		 * tweens are being run at once.
		 * </p>
		 * 
		 * @param closure	A reference to a callback function
		 * @param type		Any GoEvent type constant, the default is COMPLETE.
		 * @see removeCallback
		 * @see org.fuseproject.go.events.GoEvent#
		 */
		public function addCallback(closure : Function, type : String=GoEvent.COMPLETE):void {
			if (!_callbacks) 
				_callbacks = new Object();
			if (!_callbacks[ type ]) 
				_callbacks[ type ] = new Array();
			var a:Array = (_callbacks[ type ] as Array);
			if (a.indexOf(closure)==-1)
				a.push(closure);
		}
		
		/**
		 * Removes a method closure previously stored using addCallback.
		 * 
		 * @param closure	A reference to a function
		 * @param type		A GoEvent constant, default is COMPLETE.
		 * @see addCallback
		 * @see org.fuseproject.go.events.GoEvent#
		 */
		public function removeCallback(closure : Function, type : String=GoEvent.COMPLETE):void {
			var a:Array = (_callbacks[ type ] as Array);
			if (a) 
				while (a.indexOf(closure)>-1)
					a.splice(a.indexOf(closure), 1);
		}
		
		/**
		 * Required by IUpdatable. Performs core calculations on GoEngine pulse.
		 * 
		 * It is strongly recommended that you subclass onUpdate instead of 
		 * this method, since onUpdate provides tween start and end flags.
		 * 
		 * @param currentTime	Clock time for the current block of updates.
		 */
		override public function update(currentTime:Number) : void 
		{
			if (!_startTime) {
				setStartTime(currentTime, true);
				// Store reused parameters which reduces object-creation.
				_updateEvent = new GoEvent( GoEvent.UPDATE );
				_easeParams = new Array(0, 0, 1, _tweenDuration); 
				if (_extraEaseParams)
					_easeParams = _easeParams.concat(_extraEaseParams);
			}
			// no updates during delay or pause.
			if (_startTime > currentTime || _state == PAUSED)
				return;
			if (_doComplexUpdate) {
				complexUpdate(currentTime);
				return;
			}
			// Set _position and determine update type
			var type:String;
			if (currentTime >= _endTime) {
				type = GoEvent.COMPLETE;
				_position = 1;
			}
			else {
				type = (_currentFrame==0 ? GoEvent.START : GoEvent.UPDATE);
				_currentFrame++;
				_easeParams[0] = (currentTime - _startTime);
				_position = easing.apply(null, _easeParams);
			}
			// Run onUpdate then dispatch events (START just before first UPDATE, COMPLETE just after last)
			onUpdate(type);
			if (type==GoEvent.START) {
				_state = PLAYING;
				dispatch(type);
			}
			dispatch(GoEvent.UPDATE);
			if (type==GoEvent.COMPLETE) {
				dispatch(type);
				stop();
			}
		}
		
		/**
		 * Flags the use of complexUpdate for special features.
		 */
		protected function setUpdateFunction() : void {
			_doComplexUpdate = (_cycles!=1 || useFrames);
		}
		
		/**
		 * TODO: Add the special features - cycles, useFrames. Update play methods to handle frames.
		 */
		protected function complexUpdate(currentTime:Number) : void 
		{
			throw new Error('complexUpdate is not yet written.');
		}
		
		// -== Protected Methods ==-
		
		/**
		 * Subclass this method (instead of the update method) for simplicity. 
		 * 
		 * <p>Use this method to manipulate targets based on the current _position 
		 * setting, which is a 0-1 multiplier precalculated to the tween's position
		 * based on its easing style and the current time in the tween. It is recommended 
		 * that you use the correctValue method on each value before you apply it to a
		 * target, which corrects NaN to 0 and rounds the value if useRounding is true.
		 * 
		 * Here's an example of what an overridden onUpdate method might contain:
		 * <pre>
		 * target[ propName ] = super.correctValue(startValue + change*_position);
		 * </pre>
		 * 
		 * @param type	A constant from the class GoEvent: START, UPDATE, or COMPLETE.
		 */
		protected function onUpdate(type : String) : void 
		{
			// Subclass this method and start to implement your tween class.
		}
		
		/**
		 * Internal setup routine used by start() and other methods.
		 * 
		 * @param time			Tween start time based on getTimer
		 * @param includeDelay	Whether to calculate the delay into the start and end times
		 */
		protected function setStartTime(time : Number, includeDelay:Boolean = false) : void 
		{
			_startTime = time;
			_tweenDuration = (_duration * 1000 * Math.max(0, timeMultiplier));
			_endTime = _startTime + _tweenDuration;
			if (includeDelay) {
				_startTime += (_delay * 1000 * Math.max(0, timeMultiplier));
				_endTime += (_delay * 1000 * Math.max(0, timeMultiplier));
			}
		}
		
		/**
		 * Internal, dispatches events and executes callbacks of any pre-verified type.
		 *  
		 * @param type	Verified in addCallback, not in this method.
		 * @see org.fuseproject.go.events.GoEvent#
		 */
		protected function dispatch(type:String):void 
		{
			if (_callbacks) { 
				var a:Array = (_callbacks[ type ] as Array);
				if (a)
					for each (var callback:Function in a)
						callback();
			}
			if (type==GoEvent.UPDATE) 
				dispatchEvent(_updateEvent); // cuts down on object-creation
			else
				dispatchEvent(new GoEvent( type ));
		}
	}
}