
package org.fuseproject.go.util {
	import flash.events.Event;
	import flash.utils.Dictionary;
	
	import org.fuseproject.go.PlayableBase;
	import org.fuseproject.go.events.GoEvent;

	/**
	 * Dispatched when the group starts.
	 * @eventType org.fuseproject.go.events.START
	 */
	[Event(name="START", type="org.fuseproject.go.events.GoEvent")]

	/**
	 * Dispatched if the group is manually stopped.
	 * @eventType org.fuseproject.go.events.STOP
	 */
	[Event(name="STOP", type="org.fuseproject.go.events.GoEvent")]

	/**
	 * Dispatched after all children have dispatched a STOP or COMPLETE event.
	 * @eventType org.fuseproject.go.events.COMPLETE
	 */
	[Event(name="COMPLETE", type="org.fuseproject.go.events.GoEvent")]

	/**
	 * Batch-play a set of items and receive an event when all of them have finished.
	 * 
	 * <p>
	 * The group listens for both STOP and COMPLETE events from its children, 
	 * either of which are counted toward group completion.
	 * </p>
	 * 
	 * @author Moses Gunesch / mosessupposes.com (c) all rights reserved
	 */
	public class PlayableGroup extends PlayableBase {
		
		// -== Public Properties ==-
		
		/**
		 * Get or set the children array. Only PlayableBase items are stored. Note that 
		 * unlike the methods <code>addChild</code> and <code>removeChild</code>, 
		 * setting this property will stop any group play currently in progress.
		 */
		public function get children():Array {
			var a:Array = [];
			for (var item:Object in _children)
				a.push(item);
			return a;
		}
		public function set children(a:Array):void {
			if (_listeners > 0)
				stop();
			for each (var item:Object in a)
				if (item is PlayableBase)
					addChild(item as PlayableBase);
		}
		
		/**
		 * Determines the number of listeners currently being monitored 
		 * by the group.
		 */
		public function get listenerCount() : uint {
			return _listeners;
		}

		// -== Protected Properties ==-
		
		protected var _children: Dictionary = new Dictionary();
		protected var _listeners: uint = 0;
		
		// -== Public Methods ==-
		
		/**
		 * Constructor.
		 * 
		 * @param items	Any number of PlayableBase items as separate arguments, 
		 * 					or a single array of them.
		 */
		public function PlayableGroup(...items) {
			super();
			if (items.length > 0)
				this.children = ((items[ 0 ] is Array) ? items[ 0 ] : items);
		}
		
		/**
		 * Searches for a child with the specified id.
		 * 
		 * @param id			The item id to search for.
		 * @param deepSearch	If child is not found in the group, this option runs a 
		 * 						recursive search on any children that are PlayableGroup.
		 * @return				The SequenceStep with the matching id.
		 */
		public function getChildByID(id:*, deepSearch:Boolean=true):PlayableBase {
			for (var item:Object in _children)
				if ((item as PlayableBase).id===id)
					return (item as PlayableBase);
			if (deepSearch) {
				for (item in _children) {
					if (item is PlayableGroup) {
						var match:PlayableBase = ((item as PlayableGroup).getChildByID(id, true));
						if (match) { return (match as PlayableBase); }
					}
				}
			}
			return null;
		}
		
		/**
		 * Adds a single PlayableBase to the children array (duplicates are rejected), and optionally
		 * changes the group's play-state to match the state of the new child.
		 * 
		 * <p>In most cases when a child is added you should 
		 * 
		 * When you add a child normally you ignore the <code>adoptChildState</code>
		 * option, 
		 * 
		 * Note that if play is in progress when a child is added it does not
		 * interrupt play and the child is monitored for completion along with others.
		 * 
		 * </p>
		 * 
		 * @param item				Any instance that extends PlayableBase (PlayableBase itself should not be used directly).
		 * @param adoptChildState	Makes this group change its play-state to match the state of the new child.
		 * @return Success.
		 */
		public function addChild(item:PlayableBase, adoptChildState:Boolean=false): Boolean {
//TODO: test this method, all cases			
			if (_children[ item ])
				return false;
			if (item.state!=_state) { // Resolve an obvious mismatched play state...
									  // Normally states are both STOPPED, so the following ugliness is rarely used.
				var primary:PlayableBase = (adoptChildState ? item : this);
				var primaryPlaying:Boolean = (primary.state==PLAYING || primary.state==PLAYING_DELAY);
				var secondary:PlayableBase = (adoptChildState ? this : item);
				var secondaryPlaying:Boolean = (secondary.state==PLAYING || secondary.state==PLAYING_DELAY);
				if ( !(primaryPlaying && secondaryPlaying) ) // Less obvious, but treat PLAYING_DELAY & PLAYING as "playing."
				{
					switch (primary.state) {
						case STOPPED:
							secondary.stop();
							break;
						case PAUSED:
							if (secondary.state==STOPPED)
								secondary.start();
							secondary.pause();
							break;
						case PLAYING: 
						case PLAYING_DELAY:
							if (secondary.state==PAUSED)
								secondary.resume();
							else if (secondary.state==STOPPED) {
								if (adoptChildState && children.length==0)
									_state = PLAYING; // Special case: This group adopts child playing state as it adds first child. 
								else
									secondary.start();
							}
							break;
					}
				}
			}
			// Saved until after possible play-state changes. Now we can base listening on this group's state.
			_children[ item ] = false;
			if (_state!=STOPPED)
				listenTo(item);
			return true;
		}
		
		/**
		 * Removes a single PlayableBase from the children array.
		 * 
		 * <p>Note that if play is in progress when a child is added it does not
		 * interrupt play and the child is monitored for completion along with
		 * others.</p>
		 * 
		 * @param item	Any instance that extends PlayableBase (PlayableBase itself should not be used directly).
		 * @return Success.
		 */
		public function removeChild(item:PlayableBase): Boolean {
//TODO: test this method.
			var v:* = _children[ item ];
			if (v===null)
				return false;
			if (v===true)
				unListenTo( item );
			delete _children[ item ];
			return true;
		}
		
		/**
		 * Test whether any child has a particular play state, based on 
		 * the int constants in the PlayableBase class.
		 * 
		 * <pre>
		 * // Example: resume a paused group
		 * if ( myGroup.anyChildHasState(GoBase.PAUSED) ) {
		 *     myGroup.resume();
		 * }
		 * </pre>
		 */
		public function anyChildHasState(state:String): Boolean {
			for (var item:Object in _children)
				if ((item as PlayableBase).state==state)
					return true;
			return false;
		}
		
		// -== IPlayable implementation ==-
		
		/**
		 * Calls start on all children. 
		 * 
		 * <p>If the group is active when this method is called, a <code>stop</code> call 
		 * is automated which will result in a GoEVent.STOP event being dispatched.</p>
		 * 
		 * @return Returns true if any child in the group starts successfully.
		 */
		override public function start() : Boolean {
			stop();
			var r:Boolean = false;
			for (var item:Object in _children) {
				var started:Boolean = (item as PlayableBase).start();
				if (started)
					listenTo(item as PlayableBase);
				r = (started || r);
			}
			if (!r) return false; // all starts failed
			super.start();
			dispatchEvent(new GoEvent( GoEvent.START));
			return true;
		}

		/**
		 * If the group is active, this method stops all child items and 
		 * dispatches a GoEvent.STOP event.
		 * 
		 * @return Returns true only if all children in the group stop successfully.
		 */
		override public function stop() : Boolean {
			if (super.stop()==false)
				return false;
			if (_listeners==0) {
				dispatchEvent(new GoEvent( GoEvent.COMPLETE ));
				return true;
			}
			var r:Boolean = true;
			for (var item:Object in _children) {
				unListenTo(item as PlayableBase);
				r = ((item as PlayableBase).stop() && r);
			}
			dispatchEvent(new GoEvent( GoEvent.STOP ));
			return r;
		}
		
		/**
		 * Calls <code>pause</code> on all children.
		 * 
		 * @return  Returns true only if all playing children in the group paused successfully
		 * 			and at least one child was paused.
		 */
		override public function pause() : Boolean {
			if (_state!= PLAYING)
				return false;
			var r:Boolean = true;
			var n:uint = 0;
			for (var item:Object in _children) {
				if ((item as PlayableBase).state > PAUSED) {
					r = ((item as PlayableBase).pause() && r);
					n++;
				}
			}
			if (n>0) _state = PAUSED; // state should reflect that at least one item was paused,
									  // while return value may indicate that not all pauses succeeded.
			return (n>0 && r);
		}
		
		/**
		 * Calls <code>resume</code> on all children.
		 * 
		 * @return	Returns true only if all paused children in the group resumed successfully
		 * 			and at least one child was resumed.
		 */
		override public function resume() : Boolean {
			if (_state!= PAUSED)
				return false;
			var r:Boolean = true;
			var n:uint = 0;
			for (var item:Object in _children) {
				if ((item as PlayableBase).state == PAUSED) {
					r = ((item as PlayableBase).resume() && r);
					n++;
				}
			}
			if (n>0) _state = PLAYING; // state should reflect that at least one item was resumed,
										// while return value may indicate that not all resumes succeeded.
			return (n>0 && r);
		}
		
		/**
		 * Calls <code>skipTo</code> on all children.
		 * 
		 * @return	Returns true only if all children in the group skipTo the position successfully
		 * 			and at least one child was affected.
		 */
		override public function skipTo(position : Number) : Boolean {
			var r:Boolean = true;
			var n:uint = 0;
			for (var item:Object in _children) {
				r = ((item as PlayableBase).skipTo(position) && r);
				n++;
			}
			_state = (r ? PLAYING : STOPPED);
			return (n>0 && r);
		}
		
		// -== Protected Methods ==-
		
		/**
		 * Internal handler for item completion.
		 * @param event		GoEvent dispatched by child item. 
		 */
		protected function onItemEnd(event:GoEvent) : void {
			unListenTo(event.target as PlayableBase);
			if (_listeners==0)
				complete();
		}
		
		/**
		 * Internal handler for group completion.
		 */
		protected function complete() : void {
			stop();
		}

		/**
		 * Internal. Listen for item completion, keeping tight track of listeners.
		 * @param item	Any instance that extends PlayableBase (PlayableBase itself should not be used directly).
		 */
		protected function listenTo(item:PlayableBase) : void {
			if (_children[ item ] === false) {
				item.addEventListener(GoEvent.STOP, onItemEnd, false, 0, true);
				item.addEventListener(GoEvent.COMPLETE, onItemEnd, false, 0, true);
				_children[ item ] = true;
				_listeners++;
			}
		}
		
		/**
		 * Internal. Stop listening for item completion.
		 * @param item	Any instance that extends PlayableBase (PlayableBase itself should not be used directly).
		 * @return Number of completion listeners remaining.
		 */
		protected function unListenTo(item:PlayableBase) : void {
			if (_children[ item ] === true) {
				item.removeEventListener(GoEvent.STOP, onItemEnd);
				item.removeEventListener(GoEvent.COMPLETE, onItemEnd);
				_children[ item ] = false;
				_listeners--;
			}
		}
	}
}