
package org.fuseproject.go.util {
	import flash.events.Event;
	import flash.utils.getQualifiedClassName;
	
	import org.fuseproject.go.PlayableBase;
	import org.fuseproject.go.events.GoEvent;
	import org.fuseproject.go.events.SequenceEvent;

	/**
	 * Dispatched when the sequence starts.
	 * @eventType org.fuseproject.go.events.START
	 */
	[Event(name="START", type="org.fuseproject.go.events.GoEvent")]

	/**
	 * Dispatched when the sequence advances to its next step.
	 * @eventType org.fuseproject.go.events.SequenceEvent.ADVANCE
	 */
	[Event(name="ADVANCE", type="org.fuseproject.go.events.SequenceEvent")]

	/**
	 * Dispatched when the sequence is manually stopped, which may also occur
	 * if one of its step instances is manually stopped outside the sequence.
	 * @eventType org.fuseproject.go.events.STOP
	 */
	[Event(name="STOP", type="org.fuseproject.go.events.GoEvent")]

	/**
	 * Dispatched when the sequence successfully finishes. (In SequenceCA this event
	 * is not fired until all custom-advanced steps have dispatched STOP or COMPLETE.)
	 * @eventType org.fuseproject.go.events.COMPLETE
	 */
	[Event(name="COMPLETE", type="org.fuseproject.go.events.GoEvent")]

	/**
	 * This base class should not be used directly, use it to build sequencing classes.
	 * 
	 * <p>When subclassing, follow the instructions in the comments of the protected
	 * methods to add a standard set of public getters and methods that work with the
	 * specific datatype of your SequenceStep subclass, if you create one. The classes
	 * Sequence and SequenceCA are good examples.</p>
	 * 
	 * @see org.fuseproject.go.util.customadvance.Sequence#
	 * @see org.fuseproject.go.util.customadvance.SequenceCA#
	 * 
	 * @author Moses Gunesch / mosessupposes.com (c) all rights reserved
	 */
	public class SequenceBase extends PlayableBase {
		
		// -== Public Properties ==-
		
		/**
		 * The number of steps in the sequence.
		 */
		public function get length(): int {
			return (_steps ? _steps.length : 0);
		}
		
		/**
		 * The current play index of the sequence, starting a 0.
		 */
		public function get playIndex(): int {
			return _index;
		}

		/**
		 * Get or set the list of SequenceStep instances that defines the sequence. 
		 * 
		 * <p>
		 * When setting this property, each item included in the array must be a 
		 * subclass of PlayableBase that dispatches STOP or COMPLETE when finished. 
		 * Each item in the array passed in is converted into a SequenceStep if it
		 * is of any other PlayableBase type, such as a GoItem or PlayableGroup. 
		 * Setting this property stops any sequence play currently in progress.
		 * </p>
		 * @see getStepAt
		 * @see getStepByID
		 * @see currentStep
		 * @see lastStep
		 */
		public function get steps():Array {
			return _steps;
		}
		public function set steps(a:Array):void {
			if (_state!=STOPPED)
				stop();
			
			while (_steps.length > 0)
				_removeStepAt(_steps.length-1);
			
			for each (var item:Object in a)
				if (item is PlayableBase)
					_addStep(item as PlayableBase);
		}
		
		// -== Protected Properties ==-
		
		protected var _index: int = 0;
		protected var _steps: Array;
		
		// -== Public Methods ==-
		
		/**
		 * Constructor.
		 * 
		 * @param items		Any number of PlayableBase-derived instances (e.g. LinearGo, PlayableGroup,
		 * 					SequenceStep) as separate arguments, or a single array of them.
		 */
		public function SequenceBase(...items) {
			super();
			var className:String = getQualifiedClassName(this);
			if (className.slice(className.lastIndexOf("::")+2) == "SequenceBase") {
				throw new Error(directUseError.replace("PlayableBase", "SequenceBase"));
			}
			_steps = new Array();
			if (items.length > 0)
				steps = ((items[ 0 ] is Array) ? items[ 0 ] : items);
		}
		
		// -== IPlayable implementation ==-
		
		/**
		 * Begins a sequence.
		 * 
		 * <p>If the group is active when this method is called, a <code>stop</code> call 
		 * is automated which will result in a GoEvent.STOP event being dispatched.</p>
		 * 
		 * @return Returns true unless there are no steps in the sequence.
		 */
		override public function start() : Boolean {
			if (_steps.length==0)
				return false;
			stop();
			_getCurrentStep().start();
			dispatchEvent(new GoEvent( GoEvent.START ));
			return super.start();
		}

		/**
		 * Stops all activity and dispatches a GoEvent.STOP event.
		 * 
		 * @return Returns true unless sequence was already stopped.
		 */
		override public function stop() : Boolean {
			if (super.stop()==false || _steps.length==0)
				return false;
			_getCurrentStep().stop();
			if (_steps.length-_index > 1)
				dispatchEvent(new GoEvent( GoEvent.STOP ));
			else
				dispatchEvent(new GoEvent( GoEvent.COMPLETE ));
			_index = 0;
			return true;
		}
		
		/**
		 * Pauses sequence play.
		 * 
		 * @return  Returns true unless sequence was unable to pause any children.
		 */
		override public function pause() : Boolean {
			var prevState:String = _state;
			if (super.pause()==false)
				return false;
			if (_getCurrentStep().pause()==false) {
				_state = prevState;
				return false;
			}
			return true;
		}
		
		/**
		 * Resumes previously-paused sequence play.
		 * 
		 * @return  Returns true unless sequence was unable to resume any children.
		 */
		override public function resume() : Boolean {
			var prevState:String = _state;
			if (super.resume()==false)
				return false;
			if (_getCurrentStep().resume()==false) {
				_state = prevState;
				return false;
			}
			return true;
		}
		
		/**
		 * Stops the current step and skips to another step by sequence index.
		 * 
		 * @return Always returns true since the index is normalized to the sequence.
		 */
		override public function skipTo(index : Number) : Boolean {
			var prevIndex:int = _index;
			_index = Math.min(_steps.length, index);
			_steps[prevIndex].stop();  // _index is updated before this call so that onStepEvent ignores the item's STOP event.
			_getCurrentStep().start();
			return true;
		}
		
		// -== Add hooks for these methods to your subclass like Sequence & SequenceCA ==-
		// These methods are broken out to allow subclasses to use exact typing for their SequenceStep class.
		
		/**
		 * Developers: Add a getter called <code>currentStep</code> to your subclass as in Sequence.
		 * 
		 * @return Developers: return the correct SequenceStep type for your subclass in your corresponding public method.
		 */
		protected function _getCurrentStep() : * {
			return (_steps.length==0 ? null : _steps[_index]);
		}
		
		/**
		 * Developers: Add a getter called <code>lastStep</code> to your subclass as in Sequence.
		 * 
		 * @return Developers: return the correct SequenceStep type for your subclass in your corresponding public method.
		 */
		protected function _getLastStep() : * {
			return (_steps.length==0 ? null : _steps[ _steps.length-1 ]);
		}
		/**
		 * Developers: Add a method called <code>getStepAt</code> to your subclass as in Sequence.
		 * 
		 * @param index	An array index starting at 0.
		 * @return		Developers: return the correct SequenceStep type for your subclass in your corresponding public method.
		 */
		protected function _getStepAt(index:int) : * {
			if (index >= _steps.length)
				return null;
			return (_steps[index] as SequenceStep);
		}
		
		/**
		 * Developers: Add a method called <code>getStepByID</code> to your subclass as in Sequence.
		 * 
		 * @param id	The step instance's id to search for.
		 * @return		Developers: return the correct SequenceStep type for your subclass in your corresponding public method.
		 */
		protected function _getStepByID(id:*) : * {
			for each (var step:SequenceStep in _steps)
				if (step.id===id)
					return step;
			return null;
		}
		
		/**
		 * Developers: Add a method called <code>addStep</code> to your subclass as in Sequence.
		 * 
		 * <p>Drop the third parameter in your subclass' addStep method. Use it to be sure
		 * the correct type of wrapper is created, as in SequenceCA.</p> 
		 * 
		 * @param item			The playable item to add to the sequence.
		 * 						
		 * @param addToLastStep	If true is passed the item is added to the last
		 * 						existing SequenceStep in the steps array. This
		 * 						option should be used with individual items that
		 * 						you want added as children to the SequenceStep.
		 * 						If there are no steps yet this option ignored and
		 * 						a new step is created.
		 * 						
		 * @param stepTypeAsClass	Type for SequenceSteps. (Do not include this parameter in subclass addStep method.)
		 * 			
		 * @return New length of the steps array.
		 */
		protected function _addStep(item:PlayableBase, addToLastStep:Boolean=false, stepTypeAsClass:*=null): int {
			if (item is SequenceStep && !addToLastStep) {
				return _steps.push(item);
			}
			if (!stepTypeAsClass)
				stepTypeAsClass = SequenceStep;
			var step:SequenceStep = (addToLastStep && _steps.length > 0
									 ? (_steps.pop() as SequenceStep)
									 : new stepTypeAsClass() as SequenceStep);
			step.addChild(item);
			return _addStepAt(_steps.length, step, stepTypeAsClass); // adds listeners
		}
		
		/**
		 * Developers: Add a method called <code>addStep</code> to your subclass as in Sequence.
		 * 
		 * <p>Drop the third parameter in your subclass' addStep method. Use it to be sure
		 * the correct type of wrapper is created, as in SequenceCA.</p> 
		 * 
		 
		 * @param index			Position in the array starting at 0, or a negative 
		 * 						index like Array.splice.
		 * 					
		 * @param item			The playable item to splice into the sequence.
		 * 						
		 * @param stepTypeAsClass	Type for SequenceSteps. (Do not include this parameter in subclass addStep method.)
		 * 
		 * @return 				New length of the steps array.
		 */
		protected function _addStepAt(index:int, item:PlayableBase, stepTypeAsClass:*=null): int {
			if (_state!=STOPPED)
				stop();
			if (!stepTypeAsClass)
				stepTypeAsClass = SequenceStep;
			var step:SequenceStep = (item is SequenceStep
									 ? item as SequenceStep
									 : new stepTypeAsClass(item) as SequenceStep);
			step.addEventListener(SequenceEvent.ADVANCE, onStepEvent, false, 0, true);
			step.addEventListener(GoEvent.STOP, onStepEvent, false, 0, true);
			_steps.splice(index, 0, step);
			return _steps.length;
		}

		/**
		 * Developers: Add a method called <code>addStep</code> to your subclass as in Sequence.
		 * 
		 * @param index	Position in the array starting at 0, or a negative 
		 * 				index like Array.splice.
		 * 					
		 * @return 		Developers: return the correct SequenceStep type for your subclass in your corresponding public method.
		 */
		protected function _removeStepAt(index:int) : * {
			if (_state!=STOPPED)
				stop();
			var step:SequenceStep = _steps.splice(index, 1) as SequenceStep;
			step.removeEventListener(SequenceEvent.ADVANCE, onStepEvent);
			step.removeEventListener(GoEvent.STOP, onStepEvent);
			return step;
		}
		
		// -== Protected Methods ==-
		
		/**
		 * Internal handler for step advance.
		 * 
		 * @param event		SequenceEvent dispatched by child item. 
		 */
		protected function onStepEvent(event : Event) : void {
			// A stop() call to the sequence results in step dispatching STOP, which would recurse here.
			if (_state==STOPPED || event.target!=_steps[_index])
				return;

			// Only occurs if the SequenceItem is manually stopped outside of this manager.
			if (event.type==GoEvent.STOP) {
				stop();
				return;
			}
			
			// Normal step advance
			if (event.type==SequenceEvent.ADVANCE) {
				if (_steps.length-_index == 1) {
					// This is a safety to be sure the Sequence cannot complete while last step is still active.
					if (_getLastStep().listenerCount > 0) {
						return;
					}
					complete();
				}
				else {
					advance();
				}
			}
		}
		
		/**
		 * Internal handler for group completion.
		 */
		protected function advance() : void {
			if (_steps.length-_index > 1) {
				_index ++; // this changes currentStep value in following code
				_getCurrentStep().start();
			}
			dispatchEvent(new SequenceEvent( SequenceEvent.ADVANCE ));
		}
		
		/**
		 * Internal handler for group completion.
		 */
		protected function complete() : void {
			_index = _steps.length - 1;
			stop();
		}
	}
}