
package org.fuseproject.go.util {
	import flash.events.Event;
	
	import org.fuseproject.go.PlayableBase;
	import org.fuseproject.go.events.GoEvent;
	import org.fuseproject.go.events.SequenceEvent;
	import org.fuseproject.go.util.SequenceBase;
	import org.fuseproject.go.util.SequenceStep;

	/**
	 * Sequence with "Custom Advance" options, in which steps can specify when they should advance.
	 * 
	 * <p>This class works like Sequence but uses the special class SequenceStepCA for its steps.
	 * SequenceStepCA has a property called <code>advance</code>. When steps advance before animation 
	 * finishes, the trailing steps are tracked so that the SequenceCA doesn't dispatch its COMPLETE
	 * event until all activity has completed.</p>
	 * 
	 * <p>Any step's advance property can be set to an instance of OnDurationComplete, OnPlayableComplete,
	 * OnEventComplete or OnConditionTrue. Each of those classes defines its own parameters and rules for
	 * when the advance occurs. For example, using OnPlayableComplete a sequence can advance after one 
	 * particular item in the step finishes, without needing to wait for all the other ones in that group
	 * to complete.</p>
	 * 
	 * <p>Additionally, you can create your own custom advance types by subclassing the SequenceAdvance 
	 * base class.</p>
	 * 
	 * @see org.fuseproject.go.util.SequenceStepCA#
	 * @see org.fuseproject.go.util.customadvance.OnConditionTrue#
	 * @see org.fuseproject.go.util.customadvance.OnDurationComplete#
	 * @see org.fuseproject.go.util.customadvance.OnEventComplete#
	 * @see org.fuseproject.go.util.customadvance.OnPlayableComplete#
	 * @see org.fuseproject.go.util.customadvance.SequenceAdvance#
	 * @see org.fuseproject.go.util.customadvance.Sequence#
	 * @see org.fuseproject.go.util.customadvance.SequenceBase#
	 * 
	 * @author Moses Gunesch / mosessupposes.com (c) all rights reserved
	 */
	public class SequenceCA extends SequenceBase {
		
		// -== Public Properties ==-
		
		// Also in super:
		// length : uint   [Read-only.]
		// playIndex : int [Read-only.]
		// steps : Array
		// start() : Boolean
		// stop() : Boolean
		// pause() : Boolean
		// resume() : Boolean
		// skipTo(index:Number) : Boolean
		
		/**
		 * Returns the currently-playing SequenceStepCA.
		 * @return The currently-playing SequenceStepCA.
		 * @see getStepAt
		 * @see getStepByID
		 * @see steps
		 * @see lastStep
		 */
		public function get currentStep() : SequenceStepCA {
			return (super._getCurrentStep());
		}
		
		/**
		 * Returns the last SequenceStepCA in the current sequence.
		 * @return The last SequenceStepCA in the current sequence.
		 * @see getStepAt
		 * @see getStepByID
		 * @see steps
		 * @see currentStep
		 */
		public function get lastStep() : SequenceStepCA {
			return (super._getLastStep());
		}
		
		// -== Protected Properties ==-
		
		protected var _trailingSteps : SequenceStep;
		
		// -== Public Methods ==-
		
		/**
		 * Constructor.
		 * 
		 * @param items		Any number of PlayableBase-derived instances (e.g. LinearGo, PlayableGroup,
		 * 					SequenceStepCA) as separate arguments, or a single array of them.
		 */
		public function SequenceCA(...items) {
			super((items[ 0 ] is Array) ? items[ 0 ] : items);
		}
		
		/**
		 * Retrieves any SequenceStepCA from the steps array.
		 * @param index		An array index starting at 0.
		 * @return			The SequenceStepCA instance at this index.
		 * @see getStepByID
		 */
		public function getStepAt(index:int) : SequenceStepCA {
			return (super._getStepAt(index) as SequenceStepCA);
		}
		
		/**
		 * Locates a step with the specified id. To search within a step for a
		 * child by id, use the step instance's <code>getChildByID</code> method.
		 *  
		 * @param id			The step instance's id to search for.
		 * @return				The SequenceStepCA with the matching id.
		 */
		public function getStepByID(id:*) : SequenceStepCA {
			return (super._getStepByID(id) as SequenceStepCA);
		}
		
		/**
		 * Adds a single PlayableBase-derived instance (e.g. LinearGo, 
		 * PlayableGroup, SequenceStepCA) to the end of the steps array,
		 * or optionally adds the instance into the last SequenceStepCA
		 * instead of adding it as a new step.
		 * 
		 * <p>To remove a step use the <code>removeStepAt</code> method.</p>
		 * 
		 * @param item			The playable item to add to the sequence. Note
		 * 						that when new steps are added, any PlayableBase
		 * 						instance of a type other than SequenceStepCA is 
		 * 						automatically wrapped in a new SequenceStepCA.
		 * 						
		 * @param addToLastStep	If true is passed the item is added to the last
		 * 						existing SequenceStepCA in the steps array. This
		 * 						option should be used with individual items that
		 * 						you want added as children to the SequenceStepCA.
		 * 						If there are no steps yet this option ignored and
		 * 						a new step is created.
		 * 						
		 * @return New length of the steps array.
		 */
		public function addStep(item:PlayableBase, addToLastStep:Boolean=false): int {
			return (super._addStep(item, addToLastStep, SequenceStepCA));
		}
		
		/**
		 * Adds a single PlayableBase-derived instance (e.g. LinearGo, PlayableGroup, 
		 * SequenceStepCA) at a specific index in the steps array. Calling this method 
		 * stops any sequence play currently in progress.
		 * 
		 * @param index		Position in the array starting at 0, or a negative 
		 * 					index like Array.splice.
		 * 					
		 * @param item		The playable item to splice into the sequence.
		 * 
		 * @return 			New length of the steps array.
		 */
		public function addStepAt(index:int, item:PlayableBase): int {
			return (super._addStepAt(index, item, SequenceStepCA));
		}

		/**
		 * Removes and returns the SequenceStepCA at a specific index from the steps 
		 * array. Calling this method stops any sequence play currently in progress.
		 * 
		 * @param index		Position in the array starting at 0, or a negative 
		 * 					index like Array.splice.
		 * 					
		 * @return 			The SequenceStepCA instance removed from the steps array.
		 */
		public function removeStepAt(index:int): SequenceStepCA {
			return (super._removeStepAt(index) as SequenceStepCA);
		}
				
		// -== IPlayable implementation ==-

		/**
		 * Begins a sequence.
		 * 
		 * <p>If the group is active when this method is called, a <code>stop</code> call 
		 * is automated which will result in a GoEvent.STOP event being dispatched.</p>
		 * 
		 * @return Returns true unless there are no steps in the sequence.
		 */
		override public function start() : Boolean {
			return super.start();
		}
		
		/**
		 * Stops all activity and dispatches a GoEvent.STOP event.
		 * 
		 * @return Returns true unless sequence was already stopped.
		 */
		override public function stop() : Boolean {
			if (super.stop()==false)
				return false;
			initTrailingSteps(false);
			return true;
		}
		
		/**
		 * Pauses sequence play.
		 * 
		 * @return  Returns true unless sequence was unable to pause any children.
		 */
		override public function pause() : Boolean {
			var success:Boolean = super.pause();
			if (_trailingSteps)
				return (success || _trailingSteps.pause());
			return success;
		}
		
		/**
		 * Resumes previously-paused sequence play.
		 * 
		 * @return  Returns true unless sequence was unable to resume any children.
		 */
		override public function resume() : Boolean {
			var success:Boolean = super.resume();
			if (_trailingSteps)
				return (success || _trailingSteps.resume());
			return success;
		}
		
		/**
		 * Stops current activity and skips to another step by sequence index.
		 * 
		 * @return Always returns true since the index is normalized to the sequence.
		 */
		override public function skipTo(index : Number) : Boolean {
			// TODO: find out if stopping _trailingSteps can result in unwanted event handling in this class  
			initTrailingSteps(false);
			return super.skipTo(index);
		}
		
		// -== Protected Methods ==-
		
		/**
		 * Internal handler for item completion.
		 * @param event		SequenceEvent dispatched by child item. 
		 */
		override protected function onStepEvent(event : Event) : void {
			// A stop() call to the sequence results in step dispatching STOP, which would recurse here.
			if (_state==STOPPED)
				return;
			
			// trailing item
			if (_trailingSteps && event.target==_trailingSteps && event.type==SequenceEvent.ADVANCE) {
				initTrailingSteps(false);
				if (_steps.length-_index==1) {
					if (lastStep.state==STOPPED) {
						// A completed sequence was waiting for trailing steps to finish.
						// Otherwise, trailing items have finished before sequence ended so no action should be taken.
						complete();
					}
					else {
						// Special case where advance already fired but trailing steps have all completed: use COMPLETE
						lastStep.addEventListener(GoEvent.COMPLETE, onStepEvent);
					}
				}
				return;
			}
			
			// Finishes special case in trailing item block.
			if (event.type==GoEvent.COMPLETE) {
				lastStep.removeEventListener(GoEvent.COMPLETE, onStepEvent);
				complete();
				return;
			}
			
			super.onStepEvent(event);
		}
						
		/**
		 * Internal handler for step advance.
		 */
		override protected function advance() : void {
			if (currentStep.listenerCount > 0) {
				initTrailingSteps(true);
				_trailingSteps.addChild(currentStep, true); // 2nd param is adoptChildState flag: avoids a start call on the group.
			}
			super.advance();
		}
		
		/**
		 * Internal handler for group completion.
		 */
		override protected function complete() : void {
			_index = _steps.length - 1;
			if (!_trailingSteps)
				stop();
		}
		
		/**
		 * Internal setup for tracking items that are continuing to run after a custom advance.
		 * @param active	Whether to create or destroy the trailing-steps group.
		 */
		protected function initTrailingSteps(active:Boolean):void {
			if (!_trailingSteps && active) {
				_trailingSteps = new SequenceStep();
				_trailingSteps.id += "(_trailingSteps for sequence:"+id+")";
				_trailingSteps.addEventListener(SequenceEvent.ADVANCE, onStepEvent);
			}
			else if (!active && _trailingSteps) {
				_trailingSteps.removeEventListener(SequenceEvent.ADVANCE, onStepEvent);
				if (_state!=STOPPED)
					 _trailingSteps.stop();
				_trailingSteps = null;
			}
		}
	}
}